// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/event/room/state/state_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:meta/meta.dart';

const String _roomMemberType = 'm.room.member';

abstract class MemberChangeEvent extends StateEvent {
  static const String type = _roomMemberType;

  @override
  final String matrixType = type;

  @override
  final MemberChange content;

  MemberChangeEvent(
      this.content, RoomEventArgs args)
      : super(args, stateKey: content.subjectId.toString());

  factory MemberChangeEvent.fromJson(RoomEventArgs args, String stateKey, Map<String, dynamic> json) {
    final membershipType = json['content']['membership'];
    final subject = UserId(stateKey);

    final prevContent = json.containsKey('unsigned')
                          && json['unsigned'] != null
                       ? json['unsigned']['prev_content'] : null;

    final prevDisplayName = prevContent != null
                          ? prevContent['displayname'] : null;

    final displayName = json['content']['displayname'] ?? prevDisplayName;

    var prevAvatarUrl = prevContent != null
                        ? prevContent['avatar_url'] : null;
    if (prevAvatarUrl != null) {
      prevAvatarUrl = Uri.parse(prevAvatarUrl);
    }

    var avatarUrl = json['content']['avatar_url'] ?? prevAvatarUrl;

    if (avatarUrl != null && avatarUrl is String) {
      avatarUrl = Uri.parse(avatarUrl);
    }

    if (prevContent != null
      && prevContent['membership'] == 'join'
      && membershipType == 'join') {
      if (displayName != prevDisplayName) {
        return DisplayNameChangeEvent(subject, avatarUrl, args,
          oldName: prevDisplayName,
          newName: displayName
        );
      }

      if (avatarUrl != prevAvatarUrl) {
        return AvatarUrlChangeEvent(subject, displayName, args,
          oldUrl: prevAvatarUrl,
          newUrl: avatarUrl
        );
      }
    }

    switch (membershipType) {
      case JoinEvent.type:
        return JoinEvent(subject, displayName, avatarUrl, args);
      case LeaveEvent.type:
        return LeaveEvent(subject, displayName, avatarUrl, args);
      case BanEvent.type:
        return BanEvent(subject, displayName, avatarUrl, args);
      case InviteEvent.type:
        return InviteEvent(subject, displayName, avatarUrl, args);
      case KnockEvent.type:
      default:
        return KnockEvent(subject, displayName, avatarUrl, args);
    }
  }
}

class MemberChange extends StateEventContent {
  @override
  final String matrixType = _roomMemberType;

  @override
  final String stateKey;

  /// The [UserId] of who has joined, left, was banned, etc.
  final UserId subjectId;

  /// Display name of the [subjectId]. Might be null.
  final String displayName;

  final Uri avatarUrl;

  @protected
  final String membershipType;

  MemberChange(
    {
      @required this.subjectId,
      @required this.membershipType,
      this.displayName,
      this.avatarUrl
    })
    : stateKey = subjectId.toString();

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'membership': membershipType,
      'displayname': displayName,
      'avatar_url': avatarUrl.toString()
    });
}

class JoinEvent extends MemberChangeEvent {
  static const String type = 'join';

  JoinEvent(
      UserId subjectId, String displayName, Uri avatarUrl, RoomEventArgs args)
      : super(MemberChange(
        subjectId: subjectId,
        membershipType: type,
        displayName: displayName,
        avatarUrl: avatarUrl,
      ), args);
}

class LeaveEvent extends MemberChangeEvent {
  static const String type = 'leave';

  LeaveEvent(
    UserId subjectId, String displayName, Uri avatarUrl, RoomEventArgs args)
      : super(MemberChange(
        subjectId: subjectId,
        membershipType: type,
        displayName: displayName,
        avatarUrl: avatarUrl,
      ),
      args);
}

class BanEvent extends MemberChangeEvent {
  static const String type = 'ban';

  BanEvent(UserId subjectId, String displayName,
      Uri avatarUrl, RoomEventArgs args)
      : super(MemberChange(
      subjectId: subjectId,
      membershipType: type,
      displayName: displayName,
      avatarUrl: avatarUrl,
  ), args);
}

// TODO: Add RoomState
class InviteEvent extends MemberChangeEvent {
  static const String type = 'invite';

  InviteEvent(
    UserId subjectId, String displayName, Uri avatarUrl, RoomEventArgs args)
      : super(MemberChange(
      subjectId: subjectId,
      membershipType: type,
      displayName: displayName,
      avatarUrl: avatarUrl,
  ), args);
}

class KnockEvent extends MemberChangeEvent {
  static const String type = 'knock';

  KnockEvent(UserId subjectId, String displayName, Uri avatarUrl, RoomEventArgs args)
      : super(MemberChange(
      subjectId: subjectId,
      membershipType: type,
      displayName: displayName,
      avatarUrl: avatarUrl,
  ), args);
}

class DisplayNameChangeEvent extends JoinEvent {
  final String oldName;
  String get newName => content.displayName;

  DisplayNameChangeEvent(
    UserId subjectId,
    Uri avatarUrl,
    RoomEventArgs args,
    {
      @required this.oldName,
      @required String newName
    }
  ) : super(subjectId, newName, avatarUrl, args);

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'unsigned': {
        'prev_content': {
          'membership': 'join',
          'displayname': oldName,
          'avatar_url': content.avatarUrl.toString()
        }
      }
    });
}

class AvatarUrlChangeEvent extends JoinEvent {
  final Uri oldUrl;
  Uri get newUrl => content.avatarUrl;

  AvatarUrlChangeEvent(
    UserId subjectId,
    String displayName,
    RoomEventArgs args,
    {
      @required this.oldUrl,
      @required Uri newUrl
    }
  ) : super(subjectId, displayName, newUrl, args);

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'unsigned': {
        'prev_content': {
          'membership': 'join',
          'displayname': content.displayName,
          'avatar_url': oldUrl.toString()
        }
      }
    });
}